#!perl
package xlib;
open(COOKED,"<Xproto.cooked") || die "couldn't open Xproto.cooked: $@";



while (<COOKED>) {
    if(/PACK (\w+) ([\w\d]+)/) {
	$xpackof{$1} = $2;
	next;
    }

    if(/TYPE (\w+) (\d+) (\w+)/) {
	# simple typedef
	$xtypeof{$1} = $3;
	$xsizeof{$1} = $2;
	next;
    }
    if(/TYPE (\w+) (\d+) :(.*)/) {
	# struct
	$xtypeof{$1} = $3;
	$xsizeof{$1} = $2;
	next;
    }
    if(/DEF (\w+) (\d+)/) {
	$defines{$1} = $2;
	if(/^DEF X\_(\w+) (\d+)/) {
	    $requests[$2] = $1;
	}
	next;
    }
    print;
}


##  local(@output) = unpack($xpackof{"xConnSetupPrefix"},$reply);
##  $success = &xindexof("xConnSetupPrefix","success",@output);

sub xindexof { # type, tag, @fields
    local($typ,$tag,@fields) = @_;
    local($xele);
    local($ind) = 0;

    for (split(/\:/,$xtypeof{$typ})) {
	m/(\w+) (\w+)/;
	last if($2 eq $tag);
	$ind++;
    }
    $fields[$ind];
}

# We now have:
# given $type (ie xListFontsReq or xListFontsReply)
# $xtypeof{$type}    ":" seperated list of struct elements
# $xsizeof{$type}    number of bits in struct
# $xpackof{$type}    pack() string for struct
# $defines{$name}    anything #define'd  (ie. X_ListFonts == 49)
# $requests[$num]    name of this request (ie. 49 == ListFonts)

sub mkport {
  local($saddr,$port,*FD) = @_;
  
  local($sockaddr,$sin);
  
  $sockaddr = 'S n a4 x8';
  $sin = pack($sockaddr, 2, $port, $saddr);

  socket(FD, 2, 1, 0) || return "socket:$!";

  connect(FD, $sin) || return "connect:$!";

  select(FD); $| = 1; select(STDOUT); $| = 1;
  return "OK";
}

sub x_open_display {
  local($displayname) = @_;
#  local(XFD);
  local($reply);
  
  ($node,$off)=split(':',$displayname);
  ($name, $aliases, $type, $len, $thisaddr) = gethostbyname($node);

# network byte order is MSB first
# byte "B" for MSB first, then CARD16 major, CARD16 minor,
#  STRING8 auth prot name, STRING8 auth prot data
  $status = &mkport($thisaddr,6000+$off,XFD);
  if($status ne "OK") { return $status; }
  
  print XFD pack($xpackof{"xConnClientPrefix"},
		 unpack("S","lB")&255, 0,
		 11, 0,
		 0, 0);

  read(XFD,$reply,1024) || return "read:$!";

  local(@output) = unpack($xpackof{"xConnSetupPrefix"},$reply);
  $success = &xindexof("xConnSetupPrefix","success",@output);
		 
  if($success == 0) {
      print "Proto Major: ",
      &xindexof("xConnSetupPrefix","majorVersion",@output),
      " Proto Minor: ",
      &xindexof("xConnSetupPrefix","minorVersion",@output), "\n";
      $rlen = &xindexof("xConnSetupPrefix","lengthReason",@output);
      $xlen = $xsizeof{"xConnSetupPrefix"}/8;
      print "reason len: $rlen\n";
      ($reason) = unpack("x$xlen a$rlen",$reply);
      return "open failed: $reason";
  } else {
      $xlen = $xsizeof{"xConnSetupPrefix"}/8;
      @output = unpack("x$xlen ".$xpackof{"xConnSetup"}, $reply);
      $xlen += $xsizeof{"xConnSetup"}/8;
      $vlen = &xindexof("xConnSetup","nbytesVendor",@output);
      ($vendor) = unpack("x$xlen a$vlen",$reply);
      print "Vendor: $vendor\n";
      # ignore the rest for now...
      # @output contains most of the fields...
      $resource_id_base = &xindexof("xConnSetup","ridBase",@output);
      $resource_id_mask = &xindexof("xConnSetup","ridMask",@output);
      $resource_id_ctr = 1;

      $max_request_size = &xindexof("xConnSetup","maxRequestSize",@output);

      $nroots = &xindexof("xConnSetup","numRoots",@output);
      $nformats = &xindexof("xConnSetup","numFormats",@output);
      $rootbase = &pad4($xlen+$vlen)+$xsizeof{"xPixmapFormat"}/8*$nformats;
      # just use the first root for now...
      @root0 = unpack("x$rootbase ".$xpackof{"xWindowRoot"}, $reply);
      $root_window = &xindexof("xWindowRoot","windowId",@root0);
      $root_white = &xindexof("xWindowRoot","whitePixel",@root0);
      $root_black = &xindexof("xWindowRoot","blackPixel",@root0);


      print "Root Window: $root_window\n";

      return "OK";
  }      
}

sub next_resource_id {
    $resource_id_base | ($resource_id_mask & $resource_id_ctr++);
}

sub pad4 { 4*(int(($_[0]-1)/4)+1); }

sub x_beep {
  local($percentage) = @_;
  print XFD pack($xpackof{"xBellReq"},$defines{"X_Bell"},$percentage,1);
  "OK";
}

sub x_configure {
    local($wind,$mask,$x,$y,$width,$height,$bord,$sib,$mode)=@_;
    local($pkt,$val,$lll);

    print "using: ",
    "xcv :", $xpackof{"xConfigureValue"},
    "xcwr:", $xpackof{"xConfigureWindowReq"},
    "req :", $defines{"X_ConfigureWindow"}, "\n";

    $val = pack($xpackof{"xConfigureValue"},
		$x, $y, $width, $height);
    $lll = (length($val)+($xsizeof{"xConfigureWindowReq"}/8));
    $pkt = pack($xpackof{"xConfigureWindowReq"},
		$defines{"X_ConfigureWindow"},0,
		$lll/4,
		$wind,$mask,0);
    $pkt .= $val;
    print "len: $lll\n";
    printf ("%02x " x length($pkt), unpack("C$lll",$pkt));
    print "\n";
    print XFD $pkt;
    "OK";
}

sub x_clear_area {
    local($wind, $expose, $x,$y, $width, $height)=@_;
    $pkt = pack($xpackof{"xClearAreaReq"},
		$defines{"X_ClearArea"},
		$expose, 4, $wind, $x, $y, $width, $height);
    printf ("x clear area: ".("%02x " x length($pkt)), unpack("C*",$pkt));
    print "\n";
    print XFD $pkt;
    "OK";
}

sub x_create_pixmap {
    local($draw, $depth, $width, $height)=@_;
    local($ret);
    $pkt = pack($xpackof{"xCreatePixmapReq"},
		$defines{"X_CreatePixmap"},
		$depth, 4,
		$ret = &next_resource_id,
		$draw, $width, $height);
		
		
    printf ("x create pixmap: ".("%02x " x length($pkt)), unpack("C*",$pkt));
    print "\n";
    print XFD $pkt;
    $ret;
}

sub x_create_gc {
    local($draw, $mask, @vals)=@_;
    local($ret);
    $val = pack("L*",@vals);
    $lll = (length($val)+($xsizeof{"xCreateGCReq"}/8));
    $pkt = pack($xpackof{"xCreateGCReq"},
		$defines{"X_CreateGC"}, 0,
		$lll/4,
		$ret = &next_resource_id,
		$draw, $mask);
    $pkt .= $val;
    printf ("x create gc: ".("%02x " x length($pkt)), unpack("C*",$pkt));
    print "\n";
    print XFD $pkt;
    $ret;
}

sub x_put_image {
    local($draw,$gc,$format,$image,$destx,$desty,$width,$height,$lpad,$depth,$bpl)
	= @_;
    local($imm,$lns,$off_l,$off);
    local($chunk_lines,$chunk_bytes);
    $lns = $xh;
    $off_l = 0;
    $off = 0;
    while(length($image)-$off > 4*$xlib'max_request_size) {
	$chunk_lines = int($xlib'max_request_size/$bpl);
	$chunk_bytes = $bpl * $chunk_lines;
	$off = $bpl * $off_l;
	($imm) = unpack("x$off a$chunk_bytes",$image);
	&xlib'x_real_put_image($draw,$gc,$format,
			       $imm,$destx,$desty+$off_l,
			       $width,$chunk_lines,$lpad,$depth);
	$off_l += $chunk_lines;
    }
    $off = $bpl * $off_l;
    if (length($image) > $off) {
	($imm) = unpack("x$off a*",$image);
	&xlib'x_real_put_image($draw,$gc,$format,
			       $imm,$destx,$desty+$off_l,
			       $width,$height-$off_l,$lpad,$depth);
    }
}
sub x_real_put_image {
    local($draw,$gc,$format,$image,$destx,$desty,$width,$height,$lpad,$depth)
	= @_;

    local($ret);

    $val = $image;
    $lll = (length($val)+($xsizeof{"xPutImageReq"}/8));
    $pkt = pack($xpackof{"xPutImageReq"},
		$defines{"X_PutImage"}, $format,
		$lll/4,
		$draw, $gc, $width, $height, $destx, $desty, 
		$lpad, $depth, 0);

    $pkt .= $val;
    $pkt .= "\0" x ($lll-&pad4($lll));
    printf("x real put image: ".("%02x " x length($pkt)), unpack("C*",$pkt)) if (length($pkt)<500);
    print "\n";
    print XFD $pkt;
    "OK";
}

sub x_change_window_attributes {
# actually, we'll short cut this...
}

sub x_set_window_background_pixmap {
    local($wind,$pixmap)=@_;
    
    #BITMASK 0x1
    #PIXMAP $pixmap (long)
    $val = pack("L",$pixmap);
    $lll = (length($val)+($xsizeof{"xChangeWindowAttributesReq"}/8));
    $pkt = pack($xpackof{"xChangeWindowAttributesReq"},
		$defines{"X_ChangeWindowAttributes"},0,
		$lll/4,
		$wind,
		0x1);
    $pkt .= $val;
    printf ("xswb pixmap: ".("%02x " x length($pkt)), unpack("C$lll",$pkt));
    print "\n";
    print XFD $pkt;
    "OK";
}

1;
