#!perl

$xsizeof{"int"} = 32;
$xsizeof{"char"} = 8;
$xsizeof{"BYTE"} = 8;
$xsizeof{"BOOL"} = 8;
$xsizeof{"long"} = 32;
$xsizeof{"CARD8"} = 8;
$xsizeof{"CARD16"} = 16;
$xsizeof{"CARD32"} = 32;
$xsizeof{"INT8"} = 8;
$xsizeof{"INT16"} = 16;
$xsizeof{"INT32"} = 32;

# oops, X11R4 is missing one...
$xsizeof{"xGetModifierMappingReply"} = 256;

$xpackof{"int"} = "L";
$xpackof{"char"} = "C";
$xpackof{"BYTE"} = "C";
$xpackof{"BOOL"} = "C";
$xpackof{"long"} = "L";
$xpackof{"CARD8"} = "C";
$xpackof{"CARD16"} = "S";
$xpackof{"CARD32"} = "L";
$xpackof{"INT8"} = "C";
$xpackof{"INT16"} = "S";
$xpackof{"INT32"} = "L";



$incomment = 0;
sub fropcomment { # $incomment
    local($incomment) = @_;
    s/\/\*.*\*\///g;
    if(/\*\//) {
	$incomment = 0;
	s/^.*\*\///;
    }
    next if $incomment;
    if(/\/\*/) {
	$incomment = 1;
	s/\/\*.*//;
    }
}

sub munchproto {
    while(<XPROTO>) {
	$incomment = &fropcomment($incomment);

	if(/\#if defined\(\_\_cplusplus\) \|\| defined\(c\_plusplus\)/) {
	    while(<XPROTO>) {
		last if (/\#else/);
	    }
	}
	next if(/\#else/);
	next if(/\#endif/);
	next if(/\#undef/);
	next if(/\#ifdef NEED_EVENTS/);
	next if(/\#ifdef NEED_REPLIES/);
	next if(/\#ifndef XPROTO_H/);
	next if(/\#include/);	# maybe do these some other time.

	if(/^\#define (\w+)$/) {
	    $defines{$1} = 1;
	    print "# defined flag $1\n";
	    next;
	}
    
	if(/^\#define sz_(\w+) (\d+)$/) {
	    $xsizeof{$1} = 8*$2;
	    print "# processed size $1\n";
	    next;
	}

	if(/^\#define (\w+) CARD(\d+)$/) {
	    $xtypeof{$1} = "CARD".$2;
	    $xsizeof{$1} = $2;
	    print "# processed type $1\n";
	    next;
	}

	if(/^\#define (\w+)\s+(\d+)\s*$/) {
	    $defines{$1} = $2;
	    print "# defined flag $1 = $2\n";
	    if(/^\#define X\_/) {
		$requests[$2] = $1;
		print "# request $2 maps to $1\n";
	    }
	    next;
	}

	if(/^\#define (\w+)\s+(0x[0-9a-fA-F]+)\s*$/) {
	    $defines{$1} = oct($2);
	    print "# defined flag $1 = $2\n";
	    next;
	}
    
	if(/^typedef CARD(\d+) (\w+);/) {
	    $xtypeof{$2} = "CARD".$1;
	    $xsizeof{$2} = $1;
	    print "# processed type $2\n";
	    next;
	}

	if(/^typedef struct \{/) {
	    &process_struct();
	    next;
	}

	if(/^typedef struct _(\w+) \{/) {
	    &process_struct();
	    next;
	}

	if(/^typedef (\w+) (\w+);/) {
	    $xtypeof{$2} = $xtypeof{$1};
	    $xsizeof{$2} = $xsizeof{$1};
	    print "# processed type $2\n";
	    next;
	}

	print;
    }
}

sub process_struct {
    local($typ,$siz) = (":",0);
    while(<XPROTO>) {
	$incomment = &fropcomment($incomment);
	if(/\#if defined\(\_\_cplusplus\) \|\| defined\(c\_plusplus\)/) {
	    while(<XPROTO>) {
		last if (/\#else/);
	    }
	}
	if(/\#ifdef WORD64/) {
	    while(<XPROTO>) {
		last if (/\#else/);
	    }
	}
	next if(/\#else/);
	next if(/\#endif/);
	s/ +B16//g;
	s/ B32//g;
	last if(/\} (\w+)\;/);

	if(/union \{/) {
	    local($utt);
	    print "# handling Union\n";
	    while(<XPROTO>) {
		$utt = $_;
		if(/struct \{/) {
		    &process_struct();
		}
		last if($utt =~ /\} u\;/);
	    }
	}

	while(/(\w+)\s+(\w*)\,/) {
	    while(/(\w+)\s+(\w+)\,/) {
		$typ .= "$1 $2:";
		$siz += $xsizeof{$1};
		s/$2\,//;
	    }
	    last if (/(\w+)\s+(\w+)\;/);
	    $_ .= <XPROTO>;
	    s/\/\*.*\*\///g;
	    s/ +B16//g;
	    s/ B32//g;
	}
	if(/(\w+)\s+(\w+)\[(\d+)\]\;/) {
	    $typ .= "$1 $2 $3:";
	    $siz += $xsizeof{$1}*$3;
	}
	if(/(\w+)\s+(\w+)\;/) {
	    $typ .= "$1 $2:";
	    $siz += $xsizeof{$1};
	}

    }
    m/\} (\w+);/;
    print "# processed type $1\n";
    
    if($1 eq "xEvent") {
	# xEvent is a "special" case...
	# xsizeof is already set...
	# and we don't really need xtypeof...
	$typ = ":BYTE raw_bits 32:";
	$siz = 32 * 8;
    }
    if($siz != $xsizeof{$1}) {
	print "### mismatch:  $typ, $siz, $xsizeof{$1}\n";
    }
    $xtypeof{$1} = $typ;
    $xsizeof{$1} = $siz;
}

sub munchkeys {
    for $typ (keys %xtypeof) {
    print "Walking: $typ -- ";
	&walk_type($typ);
	$ss = 8*length(pack($xpackof{$typ},0));
    print "$xpackof{$typ}, $xsizeof{$typ} $ss\n";
	if($ss != $xsizeof{$typ}) {
	    print "MISMATCH: $xtypeof{$typ}\n";
	}
    }
}

sub walk_type {
    local($typ) = @_;
    local($xele,$save_1,$save_3);
#    print "WALK type: $xtypeof{$typ}\n";
#    print "WALK pack: $xpackof{$typ}\n";
#    print "WALK size: $xsizeof{$typ}\n";
    if($xpackof{$typ} eq "") {
#	if($xtypeof{$typ} =~ /:/) {
	    for $xele (split(/\:/,$xtypeof{$typ})) {
		next if($xele eq "");
#		print "unpacking: $xele\n";
		if($xele =~ /^(\w+) (\w+)$/) {
		    $save_1 = $1;
		    if($xpackof{$save_1} eq "") {
#			print "RECURSE: $save_1\n";
			&walk_type($save_1);
		    }
		    $xpackof{$typ} .= $xpackof{$save_1};
		} elsif($xele =~ /^(\w+) (\w+) (\d+)$/) {
		    $save_1 = $1; $save_3 = $3;
		    if($xpackof{$save_1} eq "") {
#			print "RECURSE2: $save_1\n";
			&walk_type($save_1);
		    }
		    $xpackof{$typ} .= $xpackof{$save_1}.$save_3;
		} elsif($xele =~ /^(\w+)$/) {
		    # it can't really recurse here, can it?
		    $xpackof{$typ} .= $xpackof{$1};
		} else {
		    die "bad element: $_";
		}
	    }
#	}
    }
}


open(COOKOUT,">Xproto.cooked") || die "couldn't open Xproto.cooked: $@";

open(XPROTO,"</usr/include/X11/Xproto.h") || die "Couldn't open Xproto.h: $@";
&munchproto;
close XPROTO;

open(XPROTO,"</usr/include/X11/Xprotostr.h") || die "Couldn't open Xprotostr.h: $@";
&munchproto;
close XPROTO;

open(XPROTO,"<extra-Xproto.h") || die "Couldn't open extra-Xproto.h: $@";
&munchproto;
close XPROTO;

&munchkeys;

for $xtype (sort keys %xpackof) {
    print COOKOUT "PACK $xtype $xpackof{$xtype}\n";
}
for $xtype (sort keys %xtypeof) {
    print COOKOUT "TYPE $xtype $xsizeof{$xtype} $xtypeof{$xtype}\n";
}
for $def (sort keys %defines) {
    print COOKOUT "DEF $def $defines{$def}\n";
}

close COOKOUT;

